from functools import wraps

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseBadRequest, HttpResponseForbidden


try:
    COMMANDER_KEY = settings.COMMANDER_KEY
except AttributeError as exc:
    COMMANDER_KEY = None

# don't allow the default key in production
if COMMANDER_KEY == 's00pers3cret!' and not settings.DEBUG:
    raise ImproperlyConfigured('You can\'t use the example COMMANDER_KEY; change it to something else')


def protect_with_key(view):
    @wraps(view)
    def wrapper(request, *args, **kwargs):
        if COMMANDER_KEY is None:
            return HttpResponseForbidden('Commander is disabled')
        if 'key' not in request.GET:
            return HttpResponseBadRequest('Make sure to include the ?key in your querystring')
        if request.GET['key'] != COMMANDER_KEY:
            return HttpResponseForbidden('Incorrect key')
        return view(request, *args, **kwargs)
    return wrapper



from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand, CommandError
from django.contrib.auth.management.commands import changepassword
from django.db import DEFAULT_DB_ALIAS
from django.utils.crypto import get_random_string

UserModel = get_user_model()


class Command(changepassword.Command):
    help = 'Sets a password to something random. This helps create a superuser when you can\'t log into a shell.'

    def add_arguments(self, parser):
        parser.add_argument(
            'username', nargs=1,
            help='Username to change password for.',
        )
        parser.add_argument(
            '--database', action='store', dest='database',
            default=DEFAULT_DB_ALIAS,
            help='Specifies the database to use. Default is "default".',
        )

    def handle(self, *args, **options):
        username = options['username'][0]
        try:
            u = UserModel._default_manager.using(options['database']).get(**{
                UserModel.USERNAME_FIELD: username
            })
        except UserModel.DoesNotExist:
            raise CommandError("user '%s' does not exist" % username)

        self.stdout.write("Changing password for user '%s'\n" % u)

        chars = 'abcdefghijklmnopqrstuvwxyz0123456789!@$%^*(-_)'
        password = get_random_string(30, chars)
        u.set_password(password)
        u.save()

        return "Password changed successfully for user '%s': %s" % (u, password)


from django.http import Http404
from django.shortcuts import render

from .commands import get_filtered_commands, WebRunnableCommand
from .forms import CommandForm
from .protection import protect_with_key


@protect_with_key
def command_list(request):
    commands = get_filtered_commands()
    commands_list = [ {'command': key, 'module': value} for key, value in commands.items()]

    return render(request, 'command_list.html', {
        'commands': commands_list,
        'key': request.GET['key'],
    })


@protect_with_key
def command_run(request, name):
    commands = get_filtered_commands()
    if not name in commands:
        raise Http404(name)

    runnable = WebRunnableCommand(commands[name], name)

    response_context = {
        'command_name': runnable.name,
        'command_usage': runnable.usage,
        'key': request.GET['key'],
    }

    if request.method == 'POST':
        form = CommandForm(request.POST)
        if form.is_valid():
            args = form.cleaned_data['args']
            auto_noinput = form.cleaned_data['automatic_noinput']
            response_context['command_results'] = runnable.run(args, auto_noinput)

    else:
        form = CommandForm()

    form.fields['args'].label = name
    response_context['form'] = form

    return render(request, 'command_details.html', response_context)



from django.apps import AppConfig
from django.core import checks
from django.conf import settings


def commander_key_nondefault(app_configs, **kwargs):
    errors = []
    try:
        COMMANDER_KEY = settings.COMMANDER_KEY
    except AttributeError:
        COMMANDER_KEY = None
        errors.append(
            checks.Error(
                'Missing COMMANDER_KEY',
                hint='Add COMMANDER_KEY to your settings',
                id='commander.E001'
            )
        )

    if COMMANDER_KEY == 's00pers3cret!':
        errors.append(
            checks.Warning(
                'Invalid COMMANDER_KEY',
                hint='Switch to something other than the default / example key',
                id='commander.E002'
            )
        )

    return errors


class CommanderConfig(AppConfig):
    name = 'commander'

    def ready(self):
        checks.register(checks.Tags.security)(commander_key_nondefault)



from functools import lru_cache
from io import StringIO

from django.core.management import (
    get_commands,
    load_command_class,
    call_command,
    CommandError)
from django.core.exceptions import ImproperlyConfigured


COMMAND_BLACKLIST = (
    ('shell', 'django.core'),
    ('dbshell', 'django.core'),
    ('runserver', 'django.contrib.staticfiles'),
    ('testserver', 'django.core'),
)


def get_filtered_commands():
    commands = get_commands()
    for filter_command, filter_module in COMMAND_BLACKLIST:
        if filter_command in commands and commands[filter_command] == filter_module:
            del commands[filter_command]
    return commands


class WebRunnableCommand:
    def __init__(self, app, name):
        self._app = app
        self._name = name
        self._cmd = load_command_class(app, name)
        self._parser = self._cmd.create_parser('manage.py', name)

    @property
    def app(self):
        return self._app

    @property
    def name(self):
        return self._name

    @property
    @lru_cache(maxsize=4)
    def usage(self):
        return self._parser.format_help()

    def run(self, argv, auto_noinput):
        args = argv.split()
        stdout = StringIO()
        response = ''

        has_noinput = True if '--noinput' in self.usage else False
        if has_noinput and auto_noinput:
            args.append('--noinput')

        try:
            ret_val = call_command(self._cmd, stdout=stdout, stderr=stdout, *args)
            response = "\n".join([
                f"Return value: {ret_val or None}",
                "-------",
                stdout.getvalue(),
            ])
        except (CommandError, ImproperlyConfigured) as exc:
            response = "\n".join([
                'Exception: ' + exc.__class__.__name__,
                str(exc),
            ])
        except SystemExit:
            pass
        stdout.close()

        return response



from django.urls import path

from commander import views as cmdr


commander_urls = ([
    path('', cmdr.command_list, name='list'),
    path('<str:name>/', cmdr.command_run, name='run'),
], 'commander', 'commander')


import setuptools

with open("../README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='django-commander',
    version='0.1.1',
    description='Run Django manage.py commands from the web',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Matt Cooper',
    author_email='vtbassmatt@gmail.com',
    url='https://github.com/vtbassmatt/django-commander',
    packages=setuptools.find_packages(),
    package_data={
        'commander': ['templates/*.html', 'management/commands/*.py'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ]
)